open Tokens
open Ast

let binop_of_token = function
    ADD -> Add
  | SUB -> Sub
  | MUL -> Mul
  | DIV -> Div
  | MOD -> Mod
  | EQ -> Eq
  | NEQ -> Neq
  | LT -> Lt
  | LE -> Le
  | GT -> Gt
  | GE -> Ge
  | OR -> Or
  | _ -> failwith "Not a binary operator"

let unop_of_token = function
    SUB -> Neg
  | NOT -> Not
  | _ -> failwith "Not a unary operator"

let is_expression_start = function
    IDENT(_) | NUMBER(_) | STRING(_) 
  | TRUE | FALSE | NULL | LPAREN | LBRACE | NEW | NOT | SUB | IF -> true
  | _ -> false

let log error = prerr_string error;prerr_newline();flush stderr

let parser input_char =
  let (get_pos,current_token,next_token) = Scanner.scanner input_char in
  let get_annotation pos = new_annotation pos in
  let error_token expected =
    Report.fail_at_pos (get_pos ()) ("syntax error\nexpected: "^(expected)^"\nfound: "^(token_representation (current_token ())))
  in
  let accept_token token =
    if current_token () = token then next_token ()
    else 
      error_token (token_representation token)
  in
  let unit_lit () = New(get_annotation (get_pos ()),"Unit",[]) in
  let ident () =
    match current_token () with
	IDENT(s) -> next_token ();s
      | _ -> accept_token (IDENT(""));"bad$"^(string_of_int !Report.errors)
  in
  let rec program () =
    let ann = get_annotation (get_pos ()) in
    let ds = declarations () in
    let e = expression () in
      accept_token EOF;
      Program(ann,ds,e)
  and declaration () =
    match current_token () with
	CLASS -> class_decl ()
      | _ -> fun_decl ()
  and declarations () =
    let rec aux ds =
      match current_token () with
	  CLASS | DEF -> 
	    let d = declaration () in
	      aux (d::ds)
	| _ -> List.rev ds
    in aux []
  and fun_decl () =
    let ann = get_annotation (get_pos ()) in
      accept_token DEF;
      let ident = ident () in
	accept_token LPAREN;
	let f = formals () in
	  accept_token RPAREN;
	  accept_token COLON;
	  let t = ttype () in
	    accept_token EQUAL;
	    let e = expression () in
	      accept_token SEMICOLON;
	      FunDef(ann,ident,f,t,e)
  and formal () =
    let ann = get_annotation (get_pos ()) in
    let ident = ident () in
      accept_token COLON;
      let typ = ttype () in
	Formal(ann,ident,typ)
  and formals () =
    let rec aux fs =
      match current_token () with
	  COMMA ->
	    next_token ();
	    let f = formal () in
	      aux (f::fs)
	| _ -> List.rev fs
    in
      match current_token () with
	  IDENT(_) -> aux [formal ()]
	| _ -> []
  and class_decl () =
    let ann = get_annotation (get_pos ()) in
      accept_token CLASS;
      let ident = ident () in
	accept_token LBRACE;
	let f = fields () in
	  accept_token RBRACE;
	  ClassDef(ann,ident,f)
  and field () =
    accept_token VAL;
    let f = formal () in
      accept_token SEMICOLON;
      f
  and fields () =
    let rec aux fs =
      match current_token () with
	  VAL ->
	    let f = field () in
	      aux (f::fs)
	| _ -> List.rev fs
    in aux []
  and ttype () =
    let ann = get_annotation (get_pos ()) in
      match current_token () with
	  INT ->
	    next_token ();
	    IntType(ann)
	| IDENT(s) ->
	    next_token ();
	    TypeIdent(ann,s)
	| _ ->
	    accept_token LPAREN;
	    let typs = ttypes () in
	      accept_token RPAREN;
	      accept_token ARROW;
	      FunType(ann,typs,ttype ())
  and ttypes () =
    let rec aux ts =
      match current_token () with
	  COMMA ->
	    next_token ();
	    let typ = ttype () in
	      aux (typ::ts)
	| _ -> List.rev ts
    in
      match current_token () with
	  INT | IDENT(_) | LPAREN -> aux [ttype ()]
	| _ -> []
  and statement () =
    let ann = get_annotation (get_pos ()) in
      match current_token () with
	  WHILE ->
	    next_token ();
	    accept_token LPAREN;
	    let e = expression () in
	      accept_token RPAREN;
	      accept_token LBRACE;
	      let sts = statements () in
		accept_token RBRACE;
		While(ann,e,sts)
	| VAR ->
	    next_token ();
	    let f = formal () in
	      accept_token EQUAL;
	      let e = expression () in
		accept_token SEMICOLON;
		VarDef(ann,f,e)
	| LET ->
	    next_token ();
	    let ident = ident () in
	      accept_token EQUAL;
	      let e = expression () in
		accept_token SEMICOLON;
		Assign(ann,ident,e)
	| _ -> 
	    accept_token DO;
	    let e = expression () in
	      accept_token SEMICOLON;
	      Do(ann,e)
  and statements () =
    let rec aux sts = 
      match current_token () with
	  WHILE | VAR | LET | DO ->
	    let st = statement () in
	      aux (st::sts)
	| _ ->  List.rev sts
    in aux []
  and expression () =
    let ann = get_annotation (get_pos ()) in
      match current_token () with
	  IF ->
	    next_token ();
	    accept_token LPAREN;
	    let cond = expression () in
	      accept_token RPAREN;
	      let _then = expression () in
	      let _else = 
		match current_token () with
		    ELSE ->
		      next_token();
		      expression () 
		  | _ -> unit_lit()
	      in 
		If(ann,cond,_then,_else)
	| _ -> or_expression ()
  and or_expression () =
    let rec aux e =
      match current_token () with
	  OR -> 
	    let ann = get_annotation (get_pos ()) in
	      next_token();
	      aux (BinOp(ann,Or,e,and_expression()))
	| _ -> e
    in aux (and_expression ())
  and and_expression () =
    let not e =
      UnOp(get_annotation (pos_of_expression e),Not,e)
    in
    let rec aux e =
      match current_token () with
	  AND -> 
	    let ann = get_annotation (get_pos ()) in
	      next_token ();
	      aux (not (BinOp(ann,Or,not e,not (cmp_expression()))))
	| _ -> e
    in aux (cmp_expression ())
  and cmp_expression () =
    let e = sum_expression () in
      match current_token () with
	  EQ | NEQ | GT | LT | GE | LE as token ->
	    let ann = get_annotation (get_pos ()) in
	      next_token ();
	      let e' = sum_expression () in
		BinOp(ann,binop_of_token token,e,e')
	| _ -> e
  and sum_expression () =
    let rec aux t =
      match current_token () with
	  ADD | SUB as token ->
	    let ann = get_annotation (get_pos ()) in
	      next_token ();
	      aux (BinOp(ann,binop_of_token token,t,term ()))
	| _ -> t
    in aux (term ())
  and term () =
    let rec aux t =
      match current_token () with
	  MUL | DIV | MOD as token ->
	    let ann = get_annotation (get_pos ()) in
	      next_token ();
	      let t' = simple_term () in
		aux (BinOp(ann,binop_of_token token,t,t'))
	| _ -> t
    in
      aux (simple_term ())
  and simple_term () = 
    match current_token () with
	SUB | NOT as token ->
	  let ann = get_annotation (get_pos ()) in
	    next_token ();
	    UnOp(ann,unop_of_token token,factor ())
      | _ -> factor ()
  and factor () =
    let rec aux f =
      match current_token () with
	  LPAREN -> 
	    let ann = get_annotation (get_pos ()) in
	      aux (Apply(ann,f,params ()))
	| DOT -> 
	    let ann = get_annotation (get_pos ()) in
	      next_token ();
	      aux (Select(ann,f,ident ()))
	| _ -> f
    in aux (simple_factor ())
  and simple_factor () =
    let ann = get_annotation (get_pos ()) in
    match current_token () with
	IDENT(s) -> 
	  next_token ();
	  Ident(ann,s)
      | NUMBER(n) ->
	  next_token ();
	  IntLit(ann,n)
      | STRING(s) ->
	  next_token ();
	  let new_list ann c tail = New(ann,"List",[IntLit(ann,Int32.of_int (Char.code c));tail]) in
	  let len = String.length s in
	  let rec aux i =
	    if i = len then NullLit(ann)
	    else
	      new_list ann (s.[i]) (aux (i+1))
	  in aux 0
      | TRUE -> 
	  next_token ();
	  IntLit(ann,Int32.one)
      | FALSE ->
	  next_token ();
	  IntLit(ann,Int32.zero)
      | NULL -> 
	  next_token ();
	  NullLit(ann)
      | LPAREN ->
	  next_token ();
	  let e = expression () in
	    accept_token RPAREN;
	    e
      | LBRACE ->
	  next_token ();
	  let sts = statements () in
	  let e = 
	    if current_token () = RBRACE then unit_lit ()
	    else expression ()
	  in
	    accept_token RBRACE;
	    Block(ann,sts,e)
      | _ ->
	  accept_token NEW;
	  let ident = ident () in
	    New(ann,ident,params ())
  and params () =
    accept_token LPAREN;
    match current_token () with
	RPAREN -> 
	  next_token ();
	  []
      | _ ->
	  let es = expressions () in
	    accept_token RPAREN;
	    es
  and expressions () = 
    let rec aux es =
      match current_token () with
	  COMMA ->
	    next_token ();
	    aux ((expression ())::es)
	| _ -> List.rev es
    in 
      if is_expression_start (current_token ()) then aux [expression ()]
      else []
  in program,declarations
      
let parse_program in_channel =
  let prelude = "class Unit { }\n class List { val head : Int; val tail : List; }\n" in
  let program,_ = parser (Scanner.input_char_of_in_channel in_channel) 
  and _,declarations = parser (Scanner.input_char_of_string prelude) 
  in 
  let Program(ann,ds,e) = program () in
    Program(ann,(declarations ())@ds,e)

let main filename =
  let in_channel = if filename = "--" then stdin else open_in filename in
    print_program no_tab (parse_program in_channel)
