open Tokens

let is_keyword,token_of_keyword = 
  let keywords = Hashtbl.create 10 in
    List.iter 
      (fun (s,t) -> Hashtbl.add keywords s t)
      ["do",DO;
       "else",ELSE;
       "false",FALSE;
       "if",IF;
       "let",LET;
       "new",NEW;
       "null",NULL;
       "true",TRUE;
       "var",VAR;
       "while",WHILE;
       "def",DEF;
       "class",CLASS;
       "val",VAL;
       "Int",INT];
    let is_keyword s = Hashtbl.mem keywords s
    and token_of_keyword s = Hashtbl.find keywords s
    in (is_keyword,token_of_keyword)

let is_whitespace = function
    None -> false
  | Some(c) -> c = ' ' || c = '\t' || c = '\012' || c = '\n'


let is_letter = function
    None -> false
  | Some(c) -> 
      match c with
	  'a'..'z' | 'A'..'Z' -> true
	| _ -> false

let is_digit = function
    None -> false
  | Some(c) -> 
      match c with
	  '0'..'9' -> true
	| _ -> false

let scanner input_char = 
  let line = ref 1 and column = ref 0 
		   and ch = ref (Some(' ')) 
		   and oldch = ref (Some(' ')) 
		   and token = ref BAD in
  let start = ref (!line,!column) in
  let read_char () =
    try
      Some(input_char ())
    with
	End_of_file -> None
  in
  let next_char () = 
    begin
      match !ch with
	  None -> ()
	| Some('\n') ->
	    column := 0;
	    incr line
	| _ -> incr column
    end;
    ch := read_char ();
    oldch := if (!oldch = Some('\r') && !ch = Some('\n')) then read_char() else !ch;
    ch := if (!oldch = Some('\r')) then Some('\n') else !oldch
  in
  let accept token = next_char();token in
  let rec next_token () = 
    if is_whitespace !ch then (next_char();next_token ())
    else 
      begin
	start := (!line,!column);
	match !ch with
	    Some('/') ->
	      begin
		next_char ();
		match !ch with
		    Some('/') ->
		      while !ch <> Some('\n') && !ch <> None do
			next_char ()
		      done;
		      next_token ()
		  | Some('*') ->
		      let level = ref 1 in
			while !level > 0 do
			  next_char ();
			  match !ch with
			      None -> Report.fail_at_pos (!start) "unclosed comment"
			    | Some('/') ->
				next_char ();
				if !ch = Some('*') then incr level
			    | Some('*') ->
				while !ch = Some('*') do
				  next_char ()
				done;
				if !ch = Some('/') then decr level
			    | _ -> ()
			done;
			next_char ();
			next_token ()
		  | _ -> DIV
	      end
	  | None -> EOF
	  | Some('(') -> accept LPAREN (* next_char (); LPAREN *)
	  | Some(')') -> accept RPAREN
	  | Some(':') -> accept COLON
	  | Some(';') -> accept SEMICOLON
	  | Some(',') -> accept COMMA
	  | Some('{') -> accept LBRACE
	  | Some('}') -> accept RBRACE
	  | Some('%') -> accept MOD
	  | Some('&') -> accept AND
	  | Some('|') -> accept OR
	  | Some('*') -> accept MUL
	  | Some('+') -> accept ADD
	  | Some('-') -> accept SUB
	  | Some('.') -> accept DOT
	  | Some('=') ->
	      next_char();
	      if !ch = Some('=') then accept EQ
	      else if !ch = Some('>') then accept ARROW
	      else EQUAL
	  | Some('<') ->
	      next_char();
	      if !ch = Some('=') then accept LE
	      else LT
	  | Some('>') ->
	      next_char();
	      if !ch = Some('=') then accept GE
	      else GT
	  | Some('!') ->
	      next_char();
	      if !ch = Some('=') then accept NEQ
	      else NOT
	  | Some('0') ->
	      accept (NUMBER(Int32.zero))
	  | c when is_digit c ->
	      let buf = Buffer.create 10 in
		Buffer.clear buf;
		while is_digit !ch do
		  let Some(c) = !ch in
		    Buffer.add_char buf c;
		    next_char()
		done;
		NUMBER (Int32.of_string (Buffer.contents buf))
	  | c when is_letter c ->
	      let buf = Buffer.create 10 in
		Buffer.clear buf;
		while is_digit !ch || is_letter !ch || !ch = Some('_') do
		  let Some(c) = !ch in
		    Buffer.add_char buf c;
		    next_char()
		done;
		let str = Buffer.contents buf in
		  if is_keyword str then token_of_keyword str
		  else IDENT str
	  | Some('"') ->
	      let buf = Buffer.create 10 in
		Buffer.clear buf;
		next_char();
		while !ch <> Some('"') do
		  if !ch = Some('\n') || !ch = None then Report.fail_at_pos (!start) "unterminated string"
		  else
		    begin
		      let Some(c) = !ch in
			Buffer.add_char buf c
		    end;
		  next_char();
		done;
		accept (STRING(Buffer.contents buf))
	  | Some(c) -> 
	      Report.error_at_pos (!start) ("invalid character: '"^(String.make 1 c)^"'");
	      accept BAD
      end
  in
  let get_token () = !token
  and next_token () = token := next_token()
  and get_pos () = !start 
  in
    next_char();
    next_token();
    get_pos,get_token,next_token

let input_char_of_in_channel in_channel = function () -> input_char in_channel

let input_char_of_string s =
  let i = ref 0 and n = String.length s in
    function () ->
      if !i < n then
	let c = s.[!i] in
	  incr i;c
      else raise End_of_file
      
let main filename = 
  let in_channel = if filename = "--" then stdin else open_in filename in
  let (get_pos,get_token,next_token) = scanner (input_char_of_in_channel in_channel) in
    while get_token () <> EOF do
      print_string (token_representation (get_token ()));
      print_newline();
      next_token()
    done

